unit QGame;

interface

uses
  SysUtils, Windows, Messages, Classes, Contnrs, 
  Graphics, Controls, Forms, Dialogs;

const
 TranspColor = $00FF00FF;
 
type
    TDirection = (dLeft, dRight, dTop, dDown);
    TDirections = Set Of TDirection;
type
  TGame = class;
  TSelection = class;
  TGamePoint = class (TGraphicControl)
  private
    FNumber: Integer;
    FX: Integer;
    FY: Integer;
    procedure SetX(Value: Integer);
    procedure SetY(Value: Integer);
  protected
    Glyph: TBitMap;
    function Game: TGame;
    procedure Paint; override;
    procedure SetNumber(Value: Integer); virtual;
  public
    constructor Create(AOwner: TGame; const AX, AY: Integer); reintroduce; 
            virtual;
    destructor Destroy; override;
    class function CanMove: Boolean; virtual;
    class function IsDoor: Boolean; virtual;
    class function Size: Integer;
    property Number: Integer read FNumber write SetNumber;
    property X: Integer read FX write SetX;
    property Y: Integer read FY write SetY;
  end;
  
  TDoor = class (TGamePoint)
  protected
    procedure Paint; override;
  public
    constructor Create(AOwner: TGame; const AX, AY: Integer); override;
    class function CanMove: Boolean; override;
    class function IsDoor: Boolean; override;
    procedure SetNumber(Value: Integer); override;
  end;
  
  TSpace = class (TGamePoint)
  public
    constructor Create(AOwner: TGame; const AX, AY: Integer); override;
    class function CanMove: Boolean; override;
  end;
  
  TBall = class (TGamePoint)
  private
    //PenColor: TColor;
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: 
            Integer); override;
    procedure Paint; override;
    procedure SetNumber(Value: Integer); overload; override;
  public
    constructor Create(AOwner: TGame; const AX, AY: Integer); override;
    destructor Destroy; override;
    function AllowDirections: TDirections;
    procedure Move(const ADirection: TDirection);
  end;
  
  TWall = class (TGamePoint)
  protected
    procedure Paint; override;
  public
    constructor Create(AOwner: TGame; const AX, AY: Integer); override;
  end;
  
  TGame = class (TCustomControl)
  private
    BackGroundColor: TColor;
    FNumBalls: Integer;
    FOnGameWin: TNotifyEvent;
    FOnNewGame: TNotifyEvent;
    FPoints: TObjectList;
    FSelection: TSelection;
    FX: Integer;
    FY: Integer;
    Glyph: TBitMap;
    function GetItem(const X, Y: Integer): TGamePoint;
    procedure SetX(Value: Integer);
    procedure SetY(Value: Integer);
  protected
    procedure Clear; virtual;
    procedure DecBalls;
    procedure DoNewGame;
    procedure IncBalls;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure About; virtual;
    procedure GameWin; virtual;
    procedure New; virtual;
    property Item[const X, Y: Integer]: TGamePoint read GetItem;
    property NumBalls: Integer read FNumBalls;
    property Points: TObjectList read FPoints write FPoints;
    property Selection: TSelection read FSelection;
    property X: Integer read FX write SetX;
    property Y: Integer read FY write SetY;
  published
    property OnGameWin: TNotifyEvent read FOnGameWin write FOnGameWin;
    property OnNewGame: TNotifyEvent read FOnNewGame write FOnNewGame;
  end;
  
  TGameFile = class (TGame)
  protected
    FFileName: string;
  public
    constructor Create(AOwner: TComponent); override;
    procedure Clear; override;
    procedure LoadFromFile(const AFileName: String);
    procedure Reload;
  end;
  
  TSimpleGame = class (TGame)
  public
    procedure New; override;
  end;
  
  TExGameFile = class (TGameFile)
  public
    constructor Create(AOwner: TComponent); override;
    procedure About; override;
    procedure New; override;
    function SelGame: Boolean; virtual;
  end;
  
  TSelection = class (TGraphicControl)
  private
    FBall: TBall;
    procedure SetBall(Value: TBall);
    procedure SetVisible(Value: Boolean);
  protected
    Glyph: TBitMap;
    procedure Notification(AComponent: TComponent; Operation: TOperation); 
            override;
    procedure Paint; override;
    property Ball: TBall read FBall write SetBall;
  public
    constructor Create(AOwner: TGame); reintroduce; virtual;
    destructor Destroy; override;
    function Game: TGame;
    procedure GoDown;
    procedure GoLeft;
    procedure GoRight;
    procedure GoTop;
    procedure Move(const ADirection: TDirection);
    procedure SetPosition;
    class function Size: Integer;
    property Visible write SetVisible;
  end;
  
TPointClass = class Of TGamePoint;

procedure Register;

function GetColor(const ANumber: Integer): TColor;

implementation

uses Math, QSelFase, QSobre;

procedure Register;
begin
end;

procedure SetTranspColor(Glyph: TBitMap);
begin
  Glyph.TransparentMode  := tmFixed;
  Glyph.TransparentColor := TranspColor;
  Glyph.Transparent := True;
end;

function GetColor(const ANumber: Integer): TColor;
begin
     Case ANumber Of
        0: Result := clNavy;
        1: Result := clRed;
        2: Result := clGreen;
        3: Result := clYellow;
        4: Result := clFuchsia;
        5: Result := clMaroon;
        6: Result := clBlue;
        7: Result := clAqua;
        8: Result := clOlive;
        9: Result := clTeal; 
        else Result := clLime;                               
     end;
end;
{
********************************** TGamePoint **********************************
}
constructor TGamePoint.Create(AOwner: TGame; const AX, AY: Integer);
begin
  inherited Create(AOwner);
  Glyph := TBitMap.Create;
  SetTranspColor(Glyph);
  AOwner.Points.Add(Self);
  Color  := clBlue;
  Width  := Size;
  Height := Size;
  X      := AX;
  Y      := AY;
  Parent := AOwner;
end;

destructor TGamePoint.Destroy;
begin
  with  Game.Points do
        Extract(Self);
  Glyph.Free;
  inherited Destroy;
end;

class function TGamePoint.CanMove: Boolean;
begin
  Result := False;
end;

function TGamePoint.Game: TGame;
begin
  Result := TGame(Owner);
end;

class function TGamePoint.IsDoor: Boolean;
begin
  Result := False;
end;

procedure TGamePoint.Paint;
begin
  inherited Paint;
  with  Canvas do begin
        Brush.Style := bsClear;
        Brush.Color := clNone;
  end;
end;

procedure TGamePoint.SetNumber(Value: Integer);
begin
  FNumber := Value;
  Color := GetColor(Value);
end;

procedure TGamePoint.SetX(Value: Integer);
begin
  FX := Value;
  Left := X * Size;
end;

procedure TGamePoint.SetY(Value: Integer);
begin
  FY := Value;
  Top := FY * Size;
end;

class function TGamePoint.Size: Integer;
begin
  Result := 20;
end;

{
************************************ TDoor *************************************
}
constructor TDoor.Create(AOwner: TGame; const AX, AY: Integer);
begin
  inherited Create(AOwner, AX, AY);
  Color := clWhite;
end;

class function TDoor.CanMove: Boolean;
begin
  Result := True;
end;

class function TDoor.IsDoor: Boolean;
begin
  Result := True;
end;

procedure TDoor.Paint;
begin
  inherited Paint;
  with  Canvas do begin
        Brush.Color := Self.Color;
        Brush.Style := bsSolid;
        FillRect(ClientRect);
        Pen.Color := clBlack;
        MoveTo(0, 0);
        LineTo(Size, Size);
        MoveTo(Size, 0);
        LineTo(0, Size);
  end;
end;

procedure TDoor.SetNumber(Value: Integer);
begin
  inherited SetNumber(Value);
  Color := GetColor(FNumber);
end;

{
************************************ TSpace ************************************
}
constructor TSpace.Create(AOwner: TGame; const AX, AY: Integer);
begin
  inherited Create(AOwner, AX, AY);
  Color := clSilver;
  raise Exception.Create('D''ont initialize');
end;

class function TSpace.CanMove: Boolean;
begin
  Result := True;
end;

{
************************************ TBall *************************************
}
constructor TBall.Create(AOwner: TGame; const AX, AY: Integer);
begin
  inherited Create(AOwner, AX, AY);
  Game.IncBalls;
end;

destructor TBall.Destroy;
begin
  Game.DecBalls;
  inherited Destroy;
end;

function TBall.AllowDirections: TDirections;
  
  function Allow(const X, Y: Integer): Boolean;
  var
   Next: TGamePoint;
  begin
    Next := Game.Item[X , Y];
    Result := (Next = Nil) OR ((Next Is TDoor) AND(Next.Number = Number));
  end;
  
begin
  if  Allow(X -1 , Y) then Include(Result, dLeft);
  if  Allow(X + 1, Y) then Include(Result, dRight);
  if  Allow(X, Y - 1) then Include(Result, dTop);
  if  Allow(X, Y + 1) then Include(Result, dDown);
end;

procedure TBall.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: 
        Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  Game.Selection.Ball := Self;
  Game.SetFocus;
end;

procedure TBall.Move(const ADirection: TDirection);
var
  I: Integer;
  
  function MoveTo(const AX, AY: Integer): Boolean;
  var
    Next: TGamePoint;
  begin
    Next := Game.Item[AX, AY];
    Result := (Next = Nil);
    if  Result then begin
        Self.X := AX;
        Self.Y := AY;
    end;
    Application.ProcessMessages;
    Sleep(50);
    if  (Next Is TDoor) AND (Next.Number = Number) then Free;
  end;
  
begin
  Case ADirection Of
       dLeft:  for i := X - 1 downto 0 do
                   if  not MoveTo(i, Y) then Break;
       dRight: for i := X + 1 to Game.X do
                   if  not MoveTo(i, Y) then Break;
       dTop:   for i := Y - 1 downto 0 do
                   if  not MoveTo(X, i) then Break;
       dDown:  for i := Y + 1 to Game.Y do
                   if  not MoveTo(X, i) then Break;
  end;
end;

procedure TBall.Paint;
begin
  inherited Paint;
  with  Canvas do begin
        Draw(0, 0, Glyph);
  end;
end;

procedure TBall.SetNumber(Value: Integer);
begin
  inherited SetNumber(Value);
  Glyph.LoadFromFile('ball.sys');
  Glyph.Canvas.CopyRect(ClientRect, Glyph.Canvas, Rect(Size * Value,0,Size * Value + Size, Size));
end;

{
************************************ TWall *************************************
}
constructor TWall.Create(AOwner: TGame; const AX, AY: Integer);
begin
  inherited Create(AOwner, AX, AY);
  //Color := clBlack;
  Glyph.LoadFromFile('wall.sys');
end;

procedure TWall.Paint;
var
  Neighbor: TGamePoint;
begin
  inherited Paint;
  with  Canvas do begin
        Draw(0, 0, Glyph);
        //Brush.Style := bsSolid;
        //Brush.Color := Self.Color;
        Pen.Color := clBlue;
        Pen.Style := psSolid;
        Pen.Width := 2;
        Neighbor := Game.Item[X-1, Y];
        if not((Neighbor <> nil) AND (NeighBor IS TWall)) then begin
           MoveTo(0,0);
           LineTo(0, Size);
        end;
  
        Neighbor := Game.Item[X+1, Y];
        if not((Neighbor <> nil) AND (NeighBor IS TWall)) then begin
           MoveTo(Size,0);
           LineTo(Size, Size);
        end;
  
        Neighbor := Game.Item[X, Y-1];
        if not((Neighbor <> nil) AND (NeighBor IS TWall)) then begin
           MoveTo(0, 0);
           LineTo(Size, 0);
        end;
  
        Neighbor := Game.Item[X, Y+1];
        if not((Neighbor <> nil) AND (NeighBor IS TWall)) then begin
           MoveTo(0, Size);
           LineTo(Size, Size);
        end;
  end;
end;

{
************************************ TGame *************************************
}
constructor TGame.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  //Color := clRed;
  FPoints := TObjectList.Create;
  Parent := TWinControl(AOwner);
  FSelection := TSelection.Create(Self);
  BackGroundColor := clSilver;
  //InputKeys := InputKeys + [ikArrows];
  Glyph := TBitMap.Create;
  Glyph.TransparentMode  := tmFixed;
  Glyph.TransparentColor := TranspColor;
  Glyph.Transparent := True;
  if FileExists(ExtractFilePath(Application.ExeName) + 'BackGround.sys') then
     Glyph.LoadFromFile(ExtractFilePath(Application.ExeName) + 'BackGround.sys');
end;

procedure TGame.About;
begin
  ShowMessage('Q by Lourival Oliveira da Silva'+#13+#10+
              'e-mail: lourival@lourival.eti.br   '+#13+#10+
              'site:   lourival@lourival.eti.br   ');
end;

procedure TGame.Clear;
begin
  FNumBalls := 0;
  while  (FPoints.Count > 0) do
       FPoints[0].Free;
  FNumBalls := 0;
  X := 0;
  Y := 0;
end;

procedure TGame.DecBalls;
begin
  Dec(FNumBalls);
end;

procedure TGame.DoNewGame;
begin
  if Assigned(FOnNewGame) then FOnNewGame(Self);
end;

procedure TGame.GameWin;
begin
  if Assigned(FOnGameWin) then FOnGameWin(Self);
end;

function TGame.GetItem(const X, Y: Integer): TGamePoint;
var
  I: Integer;
begin
  Result := Nil;
  for  i := 0 to FPoints.Count - 1 do
       if  (TGamePoint(FPoints[i]).X = X) AND
           (TGamePoint(FPoints[i]).Y = Y) then begin
           Result := TGamePoint(FPoints[i]);
           Break;
       end;
end;

procedure TGame.IncBalls;
begin
  Inc(FNumBalls);
end;

procedure TGame.KeyUp(var Key: Word; Shift: TShiftState);
var
  BA: Integer;
  I, J: Integer;
begin
  inherited KeyUp(Key, Shift);
  Case  Key Of
        VK_LEFT: Selection.GoLeft;
        VK_UP: Selection.GoTop;
        VK_DOWN: Selection.GoDown;
        VK_RIGHT: Selection.GORight;
        VK_TAB: begin
                   BA := Points.IndexOf(Selection.Ball);
                   for  i := 0 to Points.Count -1 do begin
                        if  (ssShift in Shift) then begin
                            J := (BA - 1) - i;
                            if  (J < 0) then
                                J := J + (Points.Count-1);
                        end
                        else begin
                            J := i + BA + 1;
                            if  (J > Points.Count-1) then
                                J := J - (Points.Count-1);
                        end;
  
                        if  (Points[J] Is TBall) then begin
                            Selection.Ball := TBall(Points[J]);
                            Break;
                        end;
                   end;
                 end;
  
  end;
end;

procedure TGame.New;
begin
  Clear;
  DoNewGame;
end;

procedure TGame.Paint;
begin
  with  Canvas do begin
        Draw(0,0, Glyph);
  {      Pen.Style   := psSolid;
        Pen.Color   := Self.Color;
        Pen.Width   := 2;
        Brush.Color := Self.Color;
        Brush.Style := bsSolid;
        Rectangle(ClientRect);}
  end;
end;

procedure TGame.SetX(Value: Integer);
begin
  FX := Value;
  Width := (FX + 1) * TGamePoint.Size;
end;

procedure TGame.SetY(Value: Integer);
begin
  FY := Value;
  Height := (FY + 1) * TGamePoint.Size;
end;

{
********************************** TGameFile ***********************************
}
constructor TGameFile.Create(AOwner: TComponent);
begin
  inherited;
end;

procedure TGameFile.Clear;
begin
  inherited Clear;
  FFileName := '';
end;

procedure TGameFile.LoadFromFile(const AFileName: String);
var
  S: TStrings;
  i, J: Integer;
begin
  Clear;
  S := TStringList.Create;
  S.LoadFromFile(AFileName);
  //Remover cabecalho e informaes de como concluir o jogo
  for  i := 0 to S.Count - 1 do begin
       for  j := 1 to Length(S[i]) div 2 do begin
            X := Max(X, J - 1);
            Y := Max(Y, I);
            Case  S[i][(j*2)-1] Of
                  'W': TWall.Create(Self, J-1, I).Number := StrToInt(S[i][j*2]);
                  //'S': TSpace.Create(Self,J-1, I).Number := StrToInt(S[i][j*2]);
                  'B': TBall.Create(Self, J-1, I).Number := StrToInt(S[i][j*2]);
                  'D': TDoor.Create(Self, J-1, I).Number := StrToInt(S[i][j*2]);
            end;
       end;
  end;
  FFileName := AFileName;
  S.Free;
end;

procedure TGameFile.Reload;
var
  OldFileName: string;
begin
  OldFileName := FFileName;
  Clear;
  LoadFromFile(OldFileName);
end;

{
********************************* TSimpleGame **********************************
}
procedure TSimpleGame.New;
var
  i, j: Integer;
begin
  inherited New;
  for  i := 0 to 1 do
       for  j := 0 to 7 do
            TSpace.Create(self, j, i);
  TSpace.Create(self, 6,2);
  TSpace.Create(self, 7,4);
  TBall.Create(self, 7, 2);
  TWall.Create(self, 7, 3);
end;

{
********************************* TExGameFile **********************************
}
constructor TExGameFile.Create(AOwner: TComponent);
begin
  inherited;
end;

procedure TExGameFile.About;
begin
  with TfrmSobre.Create(Self) do begin
       ShowModal;
       Free;
  end;
end;

procedure TExGameFile.New;
begin
  inherited New;
  SelGame;
end;

function TExGameFile.SelGame: Boolean;
begin
  with  TfrmSelFase.Create(Self) do begin
        Result := (ShowModal = mrOk);
        if  Result then
            LoadFromFile(Format('Fase%d.gme', [edtNumFase.Value]));
        Free;
  end;
end;

{
********************************** TSelection **********************************
}
constructor TSelection.Create(AOwner: TGame);
begin
  inherited Create(AOwner);
  Color  := clYellow;
  Width  := Size;
  Height := Size;
  Parent := AOwner;
  Glyph := TBitMap.Create;
  SetTranspColor(Glyph);
  Glyph.LoadFromFile('selection.sys');
end;

destructor TSelection.Destroy;
begin
  Glyph.Free;
  inherited Destroy;
end;

function TSelection.Game: TGame;
begin
  Result := TGame(Owner);
end;

procedure TSelection.GoDown;
begin
  if  (FBall <> Nil)And
      (dDown in FBall.AllowDirections) then begin
      Move(dDown);
      SetPosition;
  end;
end;

procedure TSelection.GoLeft;
begin
  if  (FBall <> Nil)And
      (dLeft in FBall.AllowDirections) then begin
      Move(dLeft);
      SetPosition;
  end;
end;

procedure TSelection.GoRight;
begin
  if  (FBall <> Nil)And
      (dRight in FBall.AllowDirections) then begin
      Move(dRight);
      SetPosition;
  end;
end;

procedure TSelection.GoTop;
begin
  if  (FBall <> Nil)And
      (dTop in FBall.AllowDirections) then begin
      Visible := False;
      Move(dTop);
      SetPosition;
  end;
end;

procedure TSelection.Move(const ADirection: TDirection);
var
  I: Integer;
  
  function MoveTo(const AX, AY: Integer): Boolean;
  var
    Next: TGamePoint;
  begin
    Next := Game.Item[AX, AY];
    Result := (Next = Nil);
    if  Result then begin
        Ball.X := AX;
        Ball.Y := AY;
    end;
    Application.ProcessMessages;
    Sleep(50);
    if  (Next Is TDoor) AND (Next.Number = Ball.Number) then begin
        Ball.Free;
        if  (Game.NumBalls = 0) then Game.GameWin;
    end;
  end;
  
begin
  Case ADirection Of
       dLeft:  for i := Ball.X - 1 downto 0 do
                   if  not MoveTo(i, Ball.Y) then Break;
       dRight: for i := Ball.X + 1 to Game.X do
                   if  not MoveTo(i, Ball.Y) then Break;
       dTop:   for i := Ball.Y - 1 downto 0 do
                   if  not MoveTo(Ball.X, i) then Break;
       dDown:  for i := Ball.Y + 1 to Game.Y do
                   if  not MoveTo(Ball.X, i) then Break;
  end;
end;

procedure TSelection.Notification(AComponent: TComponent; Operation: 
        TOperation);
begin
  inherited Notification(AComponent, Operation);
  if  (AComponent = FBall) AND
      (Operation = opRemove) then begin
      FBall := Nil;
      Visible := False;
  end;
end;

procedure TSelection.Paint;
var
  D: TDirections;
begin
  inherited Paint;
  //Imprime a seleo
  if  (FBall = nil) then Exit;
  D := FBall.AllowDirections;
  
  with  Canvas do begin
        Pen.Width   := 1;
        Pen.Style   := psSolid;
        Brush.Style := bsSolid;
        Pen.Color   := $0000FFFF;
        Brush.Color := $0000FF00;
  
        if  (dLeft in D) then begin
            Polygon([Point( 0, 15), Point( 0, 12), Point( 1, 12), Point( 1, 11),
                     Point( 2, 11), Point( 2, 10), Point( 4, 10), Point( 4, 17),
                     Point( 2, 17), Point( 2, 16), Point( 1, 16), Point( 1, 15)]);
        end;
  
        if  (dTop in D) then begin
            Polygon([Point(10,  4), Point(17,  4), Point(17,  2), Point(16,  2),
                     Point(16,  1), Point(15,  1), Point(15,  0), Point(12,  0),
                     Point(12,  1), Point(11,  1), Point(11,  2), Point(10,  2)]);
        end;
  
        if  (dRight in D) then begin
            Polygon([Point(23, 10), Point(23, 17), Point(25, 17), Point(25, 16),
                     Point(26, 16), Point(26, 15), Point(27, 15), Point(27, 12),
                     Point(26, 12), Point(26, 11), Point(25, 11), Point(25, 10)]);
        end;
  
        if  (dDown in D) then begin
            Polygon([Point(10, 23), Point(17, 23), Point(17, 25), Point(16, 25),
                     Point(16, 26), Point(15, 26), Point(15, 27), Point(12, 27),
                     Point(12, 26), Point(11, 26), Point(11, 25), Point(10, 25)]);
        end;
  end;
end;

procedure TSelection.SetBall(Value: TBall);
begin
  FBall := Value;
  if (FBall <> Nil) then begin
     SetPosition;
     Visible := True;
  end
  else
     Visible := False;
end;

procedure TSelection.SetPosition;
begin
  if  (FBall = Nil) then Exit;
  Left := FBall.Left - (Size - TGamePoint.Size) div 2;
  Top  := FBall.Top  - (Size - TGamePoint.Size) div 2;
  BringToFront;
  Visible := True;
end;

procedure TSelection.SetVisible(Value: Boolean);
begin
  inherited Visible := Value AND (Ball <> Nil);
end;

class function TSelection.Size: Integer;
begin
  Result := TGamePoint.Size + 8;
end;


end.
